open Safestd
open Safeunix
open Pingenv

type ipaddr = inet_addr
type ip_fixed = { ip_version : int; ip_ihl : int; ip_tos : int;
		 ip_len : int; ip_id : int; ip_flags : int;
		 ip_frag_off : int; ip_ttl: int; ip_proto : int;
		 ip_chksum : int }
type ip_hdr = { fixed : ip_fixed; ip_src : ipaddr; ip_dst : ipaddr }

type icmp_echo = { ping_type : int; ping_code : int; ping_checksum : int;
		  ping_id : int; ping_seq : int; ping_data : string }

type ping_pkt = { ip : ip_hdr; icmp : icmp_echo }

let icmp_echo_type = 8
let icmp_reply_type = 0

let ping pkt =
  if pkt.icmp.ping_type = icmp_echo_type & thishost() = pkt.ip.ip_dst then
    ip_forward { ip={ fixed=pkt.ip.fixed;
		      ip_src=pkt.ip.ip_dst; ip_dst=pkt.ip.ip_src };
		 icmp={ping_type=icmp_reply_type; ping_code=pkt.icmp.ping_code;
			ping_checksum=pkt.icmp.ping_checksum;
			ping_id=pkt.icmp.ping_id; ping_seq=pkt.icmp.ping_seq;
			ping_data=pkt.icmp.ping_data} }
  else if pkt.icmp.ping_type = icmp_reply_type &
    thishost() = pkt.ip.ip_dst then
    deliver pkt
  else ip_forward pkt
