open Safestd
open Graphics

let dec_loc = (20, 400)
let big_font = "*-times-bold-r-*-34-*"
let reg_font = "*-times-medium-r-*-24-*"

let _ = open_graph ""

let draw_dec color =
  let (decx, decy) = dec_loc in
  set_color color;
  moveto decx decy;
  set_font big_font;
  draw_string "DEC"

let draw_ieee color =
  let (decx, decy) = dec_loc in
  set_font big_font;
  let (offx, offy) = text_size "DEC " in
  let (ieeex, ieeey) = (decx + offx, decy) in
  set_color color;
  moveto ieeex ieeey;
  draw_string "IEEE"

(* port colors:
 *  disabled -> light gray
 *  listening -> gray
 *  learning -> 1/2 green
 *  forwarding -> green
 *  blocking -> black
 *)
let disable_color = red
let listening_color = rgb 127 127 127
let learning_color = cyan
let forwarding_color = green
let blocking_color = black

let port_offsets = 75
let port_y = 110
let rec label_ports n =
  if n > 0 then begin
    set_color foreground;
    set_font reg_font;
    moveto ((n-1) * port_offsets) port_y;
    draw_string (Printf.sprintf "port %d" n);
    label_ports (n-1)
  end else begin
    let (x_sz, y_sz) = text_size "forwarding" in
    let x_off = (size_x ()) - x_sz - 10 in
    moveto x_off port_y;
    set_color disable_color;
    draw_string "disabled";
    moveto x_off (port_y - y_sz);
    set_color listening_color;
    draw_string "listening";
    moveto x_off (port_y - 2*y_sz);
    set_color learning_color;
    draw_string "learning";
    moveto x_off (port_y - 3*y_sz);
    set_color forwarding_color;
    draw_string "forwarding";
    moveto x_off (port_y - 4*y_sz);
    set_color blocking_color;
    draw_string "blocking"
  end

let set_port_color port state =
  set_font reg_font;
  set_color (match state with
    0 -> disable_color
  | 1 -> listening_color
  | 2 -> learning_color
  | 3 -> forwarding_color
  | 4 -> blocking_color);
  let (tmp_x, _) = text_size "port 9" in
  let center_x = tmp_x/2 + (port-1)*port_offsets in
  fill_circle center_x (port_y - 30) 20

(* DEC/IEEE colors:
 * running -> green
 * stunned -> grey
 * dead -> white
 *)

let set_dec_run () = draw_dec green
let set_dec_stun () = draw_dec (rgb 100 100 100)
let set_dec_dead () = draw_dec background

let set_ieee_run () = draw_ieee green
let set_ieee_stun () = draw_ieee (rgb 100 100 100)
let set_ieee_dead () = draw_ieee background

