(* ocamlc -c -nopervasives ispan.ml *)

open Safeloader
open Safeunix
open Unixnet
open Safethread
open Log
open Bob (* formatting stuff *)
open Timer
open Config

(* *********************************************************************** *)
(* Broadcast : 01:80:c2:00:00:00
   Bob : 00:00:c0:80:e8:e7
   Horsemen (eth0) : 00:00:c0:7f:e8:e7
   Horsemen (eth1) : 00:00:c0:7e:e7:e7
   Switchlet : 00:00:c0:9b:7d:e7
*)
(* *********************************************************************** *)

let zero = 0
let one = 1

let disabled = 0
let listening = 1
let learning = 2
let forwarding = 3
let blocking = 4

let ieee_code=0x00
let ieee_config_bpdu_type = 0
let ieee_tcn_bpdu_type = 128

let no_of_ports = ref 0
let all_ports = 4
let default_path_cost = 10
let message_age_increment = 1
let no_port = 0

(* will be initializated by init_all_outs_array *)
let all_outs_array = ref (Array.create 1 (-1, Queue.create (), Mutex.create (),
					  Condition.create ()))
;;log_msg "about to mapping\n";;
let mapping = ref (Array.create 1 (List.hd !Ldbridge.all_devs))
;;log_msg "back from mapping\n";;

let ieee_broadcast =
  {priority1=0;priority2=0x0;addr1=0x01;addr2=0x80;addr3=0xc2;addr4=0;addr5=0;
  addr6=0}

(* *********************************************************************** *)

(* Rewrite to be recursive *)
let port_data_array n =
  let wolf=Array.create n  
    { pport_id={ priority=0x07; port_no=0 }; pstate=4; ppath_cost=0x01;
      pdesignated_root={priority1=0xff;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
      pdesignated_cost=0;
      pdesignated_bridge= {priority1=0xff;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
      pdesignated_port ={ priority=0x07; port_no=0 };
      ptopology_change_acknowledge=false;
      pconfig_pending=false} in 
  for i=1 to (n-1) do
    Array.set wolf i { pport_id={ priority=0x07; port_no=i }; 
		       pstate=4; ppath_cost=0x01;
		       pdesignated_root={priority1=0xff;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
		       pdesignated_cost=0;
		       pdesignated_bridge=
		       {priority1=0xff;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
		       pdesignated_port ={ priority=0; port_no=0 };
		       ptopology_change_acknowledge=false;
		       pconfig_pending=false};
  done;
  wolf

let config_bpdu_array n =
  let wolf=Array.create n  
    { ctype = ieee_config_bpdu_type;
      croot_id = {priority1=0x81;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
      croot_path_cost =0;
      cbridge_id = {priority1=0x81;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
      cport_id={ priority=0x07; port_no=0 }; cmessage_age=0; cmax_age=0x1400; 
(* hello time SB 0x200 *)
      chello_time=0x0500; cforward_delay=0x0f00; 
      ctopology_change_acknowledgement=false; ctopology_change=false} in
  for i=1 to (n-1) do
    Array.set wolf i { ctype = ieee_config_bpdu_type;
		       croot_id = {priority1=0x81;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
		       croot_path_cost =0;
		       cbridge_id = {priority1=0x81;priority2=0x0;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x7e;addr5=0xe8;addr6=0xe7};
		       cport_id={ priority=0x07; port_no=i }; cmessage_age=0; 
		       cmax_age=0x1400;chello_time=0x0500;cforward_delay=0x0f00; 
		       ctopology_change_acknowledgement=false; 
		       ctopology_change=false};
  done;
  wolf

let tcn_bpdu_array n =
  let wolf=Array.create n { tcn_type=ieee_tcn_bpdu_type} in
  for i=1 to (n-1) do
    Array.set wolf i { tcn_type=ieee_tcn_bpdu_type};
  done;
  wolf

let dumby_func () =
  ();;

let timer_array n=
  let wolf=Array.create n { active =false; valu=0.0; started=0;
				func=dumby_func} in
  for i=1 to (n-1) do
    Array.set wolf i { active=false; func =dumby_func; started=0; valu=0.0 };
  done;
  wolf


let init_all_arrays () =
  let head = Array.create 1 (-1, Queue.create (), Mutex.create (), 
			     Condition.create ()) in
  let body = Array.of_list !Ldbridge.all_outs in
  all_outs_array := Array.append head body;

  no_of_ports := Array.length body;
  
  let head2 = Array.create 1 (List.hd !Ldbridge.all_devs) in
  let body2 = Array.of_list !Ldbridge.all_devs in
  mapping := Array.append head2 body2

  

(* *********************************************************************** *)
let bridge_m = Mutex.create()

(*   Bob : 00:00:c0:80:e8:e7 *)

let bridge_info = { 
  btype = ieee_code;
  bdesignated_root = {priority1=0x00;priority2=0x01;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x9b;addr5=0x7d;addr6=0xe7};
  broot_path_cost = 0;
  broot_port = 1;
  bmax_age = 0x1400;
  bhello_time = 0x0500;
  bforward_delay = 0x0f00;
  bbridge_type = ieee_code;
  bbridge_id = {priority1=0x00;priority2=0x01;addr1=0x0;addr2=0;addr3=0xc0;
		       addr4=0x9b;addr5=0x7d;addr6=0xe7};
  bbridge_max_age = 0x1400;
  bbridge_hello_time = 0x0200;
  bbridge_forward_delay = 0x0f00;
  btopology_change_detected = false;
  btopology_change = false;
  btopology_change_time = 0x2300; (* bridge_max_age + bridge_forward_delay *)
  bhold_time = 0x01
}  

let port_info = port_data_array all_ports
let config__bpdu = config_bpdu_array all_ports 
let tcn__bpdu = tcn_bpdu_array all_ports 
  
let hello_timer = { active=false; func=dumby_func; valu=0.0; started=0 }
let tcn_timer = { active=false; func=dumby_func; valu=0.0;started=0}
let topology_change_timer = { active=false; func=dumby_func; valu=0.0;started=0 }
let message_age_timer = timer_array all_ports 
let forward_delay_timer = timer_array all_ports 
let hold_timer = timer_array all_ports 

(* *********************************************************************** *)
let designated_port port_no =
  let pti=Array.get port_info port_no in
  ( (pti.pdesignated_bridge = bridge_info.bbridge_id)
   && 
   (pti.pdesignated_port = pti.pport_id))

let root_bridge () =
  (bridge_info.bdesignated_root = bridge_info.bbridge_id)

let generate_it s =
    let len = (String.length s)-1 in
    let str = ref "String length: " in
    let a= Printf.sprintf "%d\n" len in 

    str := (!str ^ a);
    for i=0 to (len-1) do
      let tmp = Char.code (String.get s i) in
      let stmp = Printf.sprintf "%x " tmp in
      str := (!str ^ stmp);
    done;
    str := (!str ^ "\n");
    !str



let ieee_send_config_bpdu bpdu =
  let dest = address_to_string ieee_broadcast in
  let src = address_to_string bridge_info.bbridge_id in
  let dsap = to_string 0x42 in
  let ssap = to_string 0x42 in
  let ctl = to_string 0x3 in
  let z = to_string 0x0 in (* protocol_id = 00 00; protocol version id = 00 *)
  let typ = to_string bpdu.ctype in
  let flag = ref 0 in
  if (bpdu.ctopology_change_acknowledgement) then flag := 0x80;
  if (bpdu.ctopology_change) then flag := (!flag lor 0x1);
  let flags = to_string !flag in
  let root_id = identifier_to_string bpdu.croot_id in
  let root_path_cost = int_to_string bpdu.croot_path_cost 1 in
  let bridge_id = identifier_to_string bpdu.cbridge_id in
  let p_id = port_id_to_string bpdu.cport_id in
  let message_age = int_to_string bpdu.cmessage_age 0 in
  let max_age = int_to_string bpdu.cmax_age 0 in
  let hello_time = int_to_string bpdu.chello_time 0 in
  let forward_delay = int_to_string bpdu.cforward_delay 0 in
  let pad = int_to_string 0 1 in (* 0 *) 
  let len1 = to_string 0 in
  let len2 =  to_string 38 in 
  Printf.sprintf "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s" 
    dest src len1 len2 dsap ssap ctl z z z typ flags root_id root_path_cost 
    bridge_id p_id message_age max_age hello_time forward_delay pad pad


let send_config_bpdu port_no bpdu =
  let map_aux (_, q, m, c) pkt =
    Mutex.lock m;
    Queue.add pkt q;
    Condition.signal c;
    Mutex.unlock m in  
  let s= ieee_send_config_bpdu bpdu in
  let mp = Array.get !mapping port_no in
  map_aux (Array.get !all_outs_array port_no) {len=(String.length s);
					       	addr=ADDR_PKT(1,mp.dev,0); 
					       	pkt=s;
					      	next_fn=no_next}; 
  s



let rec transmit_config port_no =
  let hold_timer_expiry port_no () =
    let ht=Array.get hold_timer port_no in (* expired *)
    ht.active <- false;
    let pti=Array.get port_info port_no in
    if (pti.pconfig_pending) then begin
      transmit_config port_no;
    end in

  let start_hold_timer port_no =
    let ht=Array.get hold_timer port_no in
    if (ht.active) then
      Timer.remove (ht.valu, ht.func);
    ht.func <- hold_timer_expiry port_no;
    ht.active <- true;
    ht.valu <- ((gettimeofday()) +. ((float bridge_info.bhold_time) /. 256.0));
    Timer.add (ht.valu, ht.func) in
  
  let ht=Array.get hold_timer port_no in
  let pti=Array.get port_info port_no in
  
  if (ht.active) then begin
    pti.pconfig_pending <- true;
  end
  else begin
    let cb=Array.get config__bpdu port_no in
    cb.ctype <- ieee_config_bpdu_type;
    cb.croot_id <- bridge_info.bdesignated_root;
    cb.croot_path_cost <- bridge_info.broot_path_cost;
    cb.cbridge_id <- bridge_info.bbridge_id;
    cb.cport_id <- pti.pport_id;
    if (root_bridge ()) then begin
      cb.cmessage_age <- 0;
    end
    else begin
      let mat=Array.get message_age_timer bridge_info.broot_port in
      cb.cmessage_age <- (time() - mat.started + message_age_increment)*0x100;
    end;
    cb.cmax_age <- bridge_info.bmax_age;
    cb.chello_time <- bridge_info.bhello_time;
    cb.cforward_delay <- bridge_info.bforward_delay;
    
    cb.ctopology_change_acknowledgement <- pti.ptopology_change_acknowledge;
    pti.ptopology_change_acknowledge <- false;
    cb.ctopology_change <- bridge_info.btopology_change;
    
    send_config_bpdu port_no cb;
    pti.pconfig_pending <- false;
    start_hold_timer port_no;
  end
    


let config_bpdu_generation () =
  for port_no=1 to !no_of_ports do
    let pti=Array.get port_info port_no in
    if ( (designated_port port_no)
       	&&
       	(pti.pstate <> disabled)) then begin
	  transmit_config port_no;
	end;
  done


(* *********************************************************************** *)
(* Expiry *)

let rec start_hello_timer () =
  let hello_timer_expiry () =
    hello_timer.active <- false; (* expired *)
    config_bpdu_generation ();
    start_hello_timer () in

if (hello_timer.active) then
  Timer.remove (hello_timer.valu, hello_timer.func);

  hello_timer.func <- hello_timer_expiry;
  hello_timer.active <- true;
  hello_timer.valu <- ((gettimeofday()) +. 
		       ((float bridge_info.bhello_time) /. 256.0));
  Timer.add (hello_timer.valu, hello_timer.func);;

let stop_hello_timer () =
  hello_timer.active <- false;
  Timer.remove (hello_timer.valu, hello_timer.func);;



let become_designated_port port_no =
    let pti=Array.get port_info port_no in
	pti.pdesignated_root <- bridge_info.bdesignated_root;
	pti.pdesignated_cost <- bridge_info.broot_path_cost;
	pti.pdesignated_bridge <- bridge_info.bbridge_id;
	pti.pdesignated_port <- pti.pport_id


let designated_port_selection () =
  for port_no=1 to !no_of_ports do
    let pti=Array.get port_info port_no in
    if ( (designated_port port_no)
      ||
     	(pti.pdesignated_root <> bridge_info.bdesignated_root)
      ||
     	( bridge_info.broot_path_cost < pti.pdesignated_cost)
      ||
     	( ( bridge_info.broot_path_cost = pti.pdesignated_cost)
	 &&
	 ( ( bridge_info.bbridge_id < pti.pdesignated_bridge)
      	|| 
	  ( ( bridge_info.bbridge_id = pti.pdesignated_bridge)
	   &&
	   (pti.pport_id <= pti.pdesignated_port)
	   )))) then begin
	     become_designated_port port_no;
	   end;
  done



let root_selection () =
  let root_port = ref no_port in
  for port_no = 1 to !no_of_ports do
    let pti=Array.get port_info port_no in
    let ri=Array.get port_info !root_port in
    
    if ( ( (not(designated_port port_no))
	  &&
	  (pti.pstate <> disabled)
	  &&
	  (pti.pdesignated_root < bridge_info.bbridge_id)
       	  )
      	&&
 	( (!root_port = no_port)
       ||
	 ( pti.pdesignated_root < ri.pdesignated_root)
       ||
	 ( (pti.pdesignated_root = ri.pdesignated_root)
	  &&
	  ( ( (pti.pdesignated_cost + pti.ppath_cost)
	     < (ri.pdesignated_cost + ri.ppath_cost)
	     )
	 ||
	   ( ( (pti.pdesignated_cost + pti.ppath_cost)
	      =
	      (ri.pdesignated_cost + ri.ppath_cost)
	      )
	    &&
	    ( ( pti.pdesignated_bridge < ri.pdesignated_bridge)
	   ||
	     ( (pti.pdesignated_bridge = ri.pdesignated_bridge)
	      &&
	      ( (pti.pdesignated_port < ri.pdesignated_port)
	     ||
	       ( (pti.pdesignated_port = ri.pdesignated_port)
	   	&&
	   	(pti.pport_id < ri.pport_id)
	 	))))))))) then begin
		  root_port := port_no;
	 	end;
  done;

  bridge_info.broot_port <- !root_port; 
  if(!root_port = no_port) then begin
    bridge_info.btype <- bridge_info.bbridge_type;
    bridge_info.bdesignated_root <- bridge_info.bbridge_id;
    bridge_info.broot_path_cost <- zero;
  end
  else begin
    let r=Array.get port_info !root_port in
    bridge_info.bdesignated_root <- r.pdesignated_root;
    bridge_info.broot_path_cost <- (r.pdesignated_cost + r.ppath_cost);
  end



let configuration_update () =
  root_selection ();
  designated_port_selection ()



let set_port_state port_no state =
    let pti=Array.get port_info port_no in
    let mp = Array.get !mapping port_no in
    pti.pstate <- state;
    mp.state <- state


let designated_for_some_port () =
    let valu = ref false in
	for port_no=1 to !no_of_ports do
      	let pti=Array.get port_info port_no in
	    if (pti.pdesignated_bridge = bridge_info.bbridge_id) then begin
	      valu := true;
	    end;
	done;
	!valu



let ieee_send_tcn_bpdu bpdu =
  let dest = address_to_string ieee_broadcast in
  let src = address_to_string bridge_info.bbridge_id in
  let dsap = to_string 0x42 in
  let ssap = to_string 0x42 in
  let ctl = to_string 0x3 in
  let z = to_string 0x0 in (* protocol_id = 00 00; protocol version id = 00 *)
  let typ = to_string bpdu.tcn_type in
  let len1 = to_string 0 in
  let len2 =  to_string 38 in (* 38 *)
  Printf.sprintf "%s%s%s%s%s%s%s%s%s%s%s%s"
    dest src len1 len2 dsap ssap ctl z z z typ (create_pad 34)


let send_tcn_bpdu port_no bpdu =
  let map_aux (_, q, m, c) pkt =
    Mutex.lock m;
    Queue.add pkt q;
    Condition.signal c;
    Mutex.unlock m in  
  let s= ieee_send_tcn_bpdu bpdu in
  let mp = Array.get !mapping port_no in
  map_aux (Array.get !all_outs_array port_no) {len=(String.length s);
					      	addr=ADDR_PKT(1,mp.dev,0);
					       	pkt=s;
					      	next_fn=no_next}; 
  s



let transmit_tcn () =
  let port_no = bridge_info.broot_port in
  let tb=Array.get tcn__bpdu port_no in
  tb.tcn_type <- ieee_tcn_bpdu_type;
  send_tcn_bpdu port_no tb



let rec start_topology_change_timer () =
  let topology_change_timer_expiry () =
    topology_change_timer.active <- false; (* expired *)
    bridge_info.btopology_change_detected <- false;
    bridge_info.btopology_change <- false in

if (topology_change_timer.active) then
  Timer.remove (topology_change_timer.valu, topology_change_timer.func);

  topology_change_timer.func <- topology_change_timer_expiry;
  topology_change_timer.active <- true;
  topology_change_timer.valu <- ((gettimeofday()) +. 
				 ((float bridge_info.btopology_change_time) 
				  /. 256.0));
  Timer.add (topology_change_timer.valu, topology_change_timer.func);;


let stop_topology_change_timer () =
  topology_change_timer.active <- false; 
  Timer.remove (topology_change_timer.valu, topology_change_timer.func);;   



let rec start_tcn_timer () =
  let tcn_timer_expiry () =
    tcn_timer.active <- false; (* expired *)
    transmit_tcn ();
    start_tcn_timer () in

if(tcn_timer.active) then
  Timer.remove (tcn_timer.valu, tcn_timer.func);
 
  tcn_timer.func <- tcn_timer_expiry;
  tcn_timer.active <- true;
  tcn_timer.valu <- ((gettimeofday()) +. 
		     ((float bridge_info.bbridge_hello_time) /. 256.0));
  Timer.add (tcn_timer.valu, tcn_timer.func);;

let stop_tcn_timer () =
  tcn_timer.active <- false;
  Timer.remove (tcn_timer.valu, tcn_timer.func);;   



let topology_change_detection () =
    if (root_bridge () ) then begin
      bridge_info.btopology_change <- true;
      start_topology_change_timer ();
    end
    else begin
      if (bridge_info.btopology_change_detected = false) then begin
	transmit_tcn ();
	start_tcn_timer ();
      end;
    end;
      bridge_info.btopology_change_detected <- true




let rec start_forward_delay_timer port_no =
  let forward_delay_timer_expiry port_no () =
    let fdt=Array.get forward_delay_timer port_no in (* expired *)
    fdt.active <- false;
    let pti=Array.get port_info port_no in
	if (pti.pstate = listening) then begin
	  set_port_state port_no learning;
	  start_forward_delay_timer port_no;
	end
	else begin
	  if (pti.pstate = learning) then begin
	    set_port_state port_no forwarding;
	    if (designated_for_some_port ()) then begin
	      topology_change_detection ();
	    end;
	  end;
	end in

  let fdt=Array.get forward_delay_timer port_no in

if (fdt.active) then
  Timer.remove (fdt.valu, fdt.func);

  fdt.active <- true;
  fdt.func <- forward_delay_timer_expiry port_no;
  fdt.valu <- ((gettimeofday()) +.((float bridge_info.bforward_delay) /. 256.0));
  Timer.add (fdt.valu, fdt.func);;


let stop_forward_delay_timer port_no =
    let fdt = Array.get forward_delay_timer port_no in
    fdt.active <- false;
    Timer.remove (fdt.valu, fdt.func);;   


let make_forwarding port_no =
  let pti=Array.get port_info port_no in
  if(pti.pstate = blocking) then begin
    set_port_state port_no listening;
    start_forward_delay_timer port_no;
  end

let make_blocking port_no =
  let pti=Array.get port_info port_no in
  if( (pti.pstate <> disabled)
     &&
     (pti.pstate <> blocking)
     ) then begin
       if ( (pti.pstate = forwarding)
	 ||
	   (pti.pstate = learning)
	   ) then begin
	     topology_change_detection ();
	   end;
       set_port_state port_no blocking;
       stop_forward_delay_timer port_no;
     end


let stop_message_age_timer port_no =
    let mat=Array.get message_age_timer port_no in
    mat.active <- false; 
    Timer.remove (mat.valu, mat.func);;   



let port_state_selection () =
   for port_no=1 to !no_of_ports do
      let pti=Array.get port_info port_no in
      if (port_no = bridge_info.broot_port) then begin
	pti.pconfig_pending <- false;
	pti.ptopology_change_acknowledge <- false;
	make_forwarding port_no;
      end
      else begin
	if (designated_port port_no) then begin
	  stop_message_age_timer port_no;
	  make_forwarding port_no;
	end
	else begin
	  pti.pconfig_pending <- false;
	  pti.ptopology_change_acknowledge <- false;
	  make_blocking port_no;
	end;
      end;
    done



let stop_message_age_timer_no_kill port_no =
    let mat=Array.get message_age_timer port_no in
    mat.active <- false; 
    Timer.remove_no_kill (mat.valu, mat.func);;   

let start_message_age_timer port_no message_age =
  let message_age_timer_expiry port_no () =
    if(Mutex.try_lock bridge_m) then begin
      stop_message_age_timer_no_kill port_no;
      let mat = Array.get message_age_timer port_no in (* expired *)
      mat.active <- false;
      let root = root_bridge () in
      become_designated_port port_no;
      configuration_update ();
      port_state_selection ();
      
      if ( (root_bridge()) && (not root)) then begin
      	bridge_info.bmax_age <- bridge_info.bbridge_max_age;
      	bridge_info.bhello_time <- bridge_info.bbridge_hello_time;
      	bridge_info.bforward_delay <- bridge_info.bbridge_forward_delay;
      	
      	topology_change_detection ();
      	stop_tcn_timer ();
      	config_bpdu_generation ();
      	start_hello_timer ();
      end;
      Mutex.unlock bridge_m;
    end in

  let mat = Array.get message_age_timer port_no in

if (mat.active) then
  Timer.remove (mat.valu, mat.func);

  mat.func <- message_age_timer_expiry port_no;
  mat.active <- true;
  mat.started <- time() - (message_age/0x100);
  mat.valu <- gettimeofday() +. 
	      float ((bridge_info.bmax_age - message_age) / 0x100);
  Timer.add (mat.valu, mat.func);;


(* *********************************************************************** *)

let stop_hold_timer port_no =
    let ht = Array.get hold_timer port_no in
    ht.active <- false;
    Timer.remove (ht.valu, ht.func);;   

(* *********************************************************************** *)

let supersedes_port_info port_no config =
  let pti=Array.get port_info port_no in
  ( (config.croot_id < pti.pdesignated_root)
  ||
    ( (config.croot_id = pti.pdesignated_root)
     && 
     ( ( config.croot_path_cost < pti.pdesignated_cost)
     || 
       ( ( config.croot_path_cost = pti.pdesignated_cost)
	&&
	( (config.cbridge_id < pti.pdesignated_bridge)
     	||
	  ( (config.cbridge_id = pti.pdesignated_bridge)
	   &&
	   ( (config.cbridge_id <> bridge_info.bbridge_id)
	   ||
	     (config.cport_id <= pti.pdesignated_port)
	     )))))))


let record_config_information port_no config =
  let pti=Array.get port_info port_no in
  pti.pdesignated_root <- config.croot_id;
  pti.pdesignated_cost <- config.croot_path_cost;
  pti.pdesignated_bridge <- config.cbridge_id;
  pti.pdesignated_port <- config.cport_id;
  start_message_age_timer port_no config.cmessage_age


let record_config_timeout_valus config =
  bridge_info.bmax_age <- config.cmax_age;
  bridge_info.bhello_time <- config.chello_time;
  bridge_info.bforward_delay <- config.cforward_delay;
  bridge_info.btopology_change <- config.ctopology_change




let reply port_no =
  transmit_config port_no



let topology_change_acknowledged () =
    bridge_info.btopology_change_detected <- false;
    stop_tcn_timer ()

let acknowledge_topology_change port_no =
    let pti=Array.get port_info port_no in
	pti.ptopology_change_acknowledge <- true;
	transmit_config port_no


(* *********************************************************************** *)
(* Operation of the Protocol (4.7) *)

let received_config_bpdu port_no config =
  Mutex.lock bridge_m;
  let root = root_bridge() in
	let pti=Array.get port_info port_no in
	    if(pti.pstate <> disabled) then begin
	      if (supersedes_port_info port_no config) then begin
		record_config_information port_no config;
	      	configuration_update ();
	      	port_state_selection ();

	      	if ( (not(root_bridge ())) && root) then begin
		  stop_hello_timer ();
		  if (bridge_info.btopology_change_detected) then begin
		    stop_topology_change_timer ();
		    transmit_tcn ();
		    start_tcn_timer ();
		  end;
		end;
		
                  if (port_no = bridge_info.broot_port) then begin
		    record_config_timeout_valus config;
		    config_bpdu_generation ();
		    if (config.ctopology_change_acknowledgement) then begin
		      topology_change_acknowledged ();
		    end;
		  end;
	      end
	      else begin
	      	if (designated_port port_no) then begin
		  reply port_no;
		end;
	      end;
	    end;
	Mutex.unlock bridge_m;;

let received_tcn_bpdu port_no tcn =
    let pti=Array.get port_info port_no in
	if (pti.pstate <> disabled ) then begin
	  if (designated_port port_no) then begin
	    topology_change_detection ();
	    acknowledge_topology_change port_no;
	  end;
	end


(* *********************************************************************** *)
(* Management of the Bridge Protocol Entity (4.8) *)

let initialize_port port_no =
  become_designated_port port_no;
  set_port_state port_no blocking;
  let pti=Array.get port_info port_no in
  pti.ptopology_change_acknowledge <- false;
  pti.pconfig_pending <- false;
  stop_message_age_timer port_no;
  stop_forward_delay_timer port_no;
  stop_hold_timer port_no

let initialization hostname =
  Timer.create (); 
  let config_info = Hashtbl.find config_hash hostname in
  bridge_info.bbridge_id <- config_info.bridge_id;
  bridge_info.bbridge_max_age <- config_info.bridge_max_age;
  bridge_info.bbridge_hello_time <- config_info.bridge_hello_time;
  bridge_info.bbridge_forward_delay <- config_info.bridge_forward_delay;

  bridge_info.bdesignated_root <- bridge_info.bbridge_id;
  bridge_info.broot_path_cost <- zero;
  bridge_info.broot_port <- no_port;
  
  bridge_info.bmax_age <- bridge_info.bbridge_max_age;
  bridge_info.bhello_time <- bridge_info.bbridge_hello_time;
  bridge_info.bforward_delay <- bridge_info.bbridge_forward_delay;
  
  bridge_info.btopology_change_detected <- false;
  bridge_info.btopology_change <- false;
  
  stop_tcn_timer ();
  stop_topology_change_timer ();

  for port_no=1 to !no_of_ports do
    initialize_port port_no;
  done;
  
  port_state_selection ();
  config_bpdu_generation ();
  start_hello_timer ();;





let enable_port port_no =
    initialize_port port_no;
    port_state_selection ()

let disable_port port_no =
    let root = root_bridge () in
	become_designated_port port_no;
	set_port_state port_no disabled;
	let pti=Array.get port_info port_no in
	    pti.ptopology_change_acknowledge <- false;
	    pti.pconfig_pending <- false;
	    stop_message_age_timer port_no;
	    stop_forward_delay_timer port_no;
	    configuration_update ();
	    port_state_selection ();
	    if ( (root_bridge ()) && (not root)) then begin
	      bridge_info.bmax_age <- bridge_info.bbridge_max_age;
	      bridge_info.bhello_time <- bridge_info.bbridge_hello_time;
	      bridge_info.bforward_delay <- bridge_info.bbridge_forward_delay;
	      topology_change_detection ();
	      stop_tcn_timer ();
	      config_bpdu_generation ();
	      start_hello_timer ();
	    end

let set_bridge_priority new_bridge_id =
    let root = root_bridge () in
	for port_no=1 to !no_of_ports do
	  if (designated_port port_no) then begin
	    let pti=Array.get port_info port_no in
		pti.pdesignated_bridge <- new_bridge_id;
	  end;
	done;

	bridge_info.bbridge_id <- new_bridge_id;
	configuration_update ();
	port_state_selection ();
	if ( (root_bridge ()) && (not root)) then begin
	  bridge_info.bmax_age <- bridge_info.bbridge_max_age;
	  bridge_info.bhello_time <- bridge_info.bbridge_hello_time;
	  bridge_info.bforward_delay <- bridge_info.bbridge_forward_delay;
	  
	  topology_change_detection ();
	  stop_tcn_timer ();
	  config_bpdu_generation ();
	  start_hello_timer ();
	end

let set_port_priority port_no new_port_id =
    let pti=Array.get port_info port_no in
	if (designated_port port_no ) then begin
	  pti.pdesignated_port <- new_port_id;
	end;
    
	  pti.pport_id <- new_port_id;
	
	if ( (bridge_info.bbridge_id = pti.pdesignated_bridge)
	    &&
	    (pti.pport_id < pti.pdesignated_port)
	    ) then begin
	      become_designated_port port_no;
	      port_state_selection ();
	    end

let set_path_cost port_no path_cost =
    let pti=Array.get port_info port_no in
	pti.ppath_cost <- path_cost;
	configuration_update ();
	port_state_selection ()







