(* log.ml
 *
 * implementation for logging messages
 *
 * $Id: log_priv.ml,v 1.2 1998/01/14 16:02:14 mwh Exp $
 * $Revision: 1.2 $
 *
 *)

open Unix

let logfile = ref None
let log_mutex = Mutex.create()

(* Function to properly enter and exit a critical section *)
let critical m thunk =
  try
    Mutex.lock m;
    let result = thunk() in
    Mutex.unlock m;
    result
  with e -> Mutex.unlock m; raise e

let set_log_file fname =
  critical log_mutex 
    (function x ->
      (match !logfile with
    	None ->
	  let fd = openfile fname [O_TRUNC; O_WRONLY; O_CREAT] 0o644 in
	  logfile := Some(fd)
      | _ -> ()))

let log_msg msg =
  critical log_mutex
    (function x ->
      let result =
    	(match (!logfile) with
	  None ->
      	    let fd = openfile "./log" [O_TRUNC; O_WRONLY; O_CREAT] 0o644 in
      	    logfile := Some(fd);
      	    ThreadUnix.write fd msg 0 (String.length msg)
    	| Some(fd) ->
      	    ThreadUnix.write fd msg 0 (String.length msg)) in
      result)

(*
 * $Log: log_priv.ml,v $
; Revision 1.2  1998/01/14  16:02:14  mwh
; Modified so that it will properly integrate with PLAN:
;   - added hook to suck up concatenated plan source files into safeloader.ml*
;   - added safe file pland which begins the process of demux'ing plan packets
;   - modified safeunix.unixnet to have function get_hw_addr
;   - added some (non-interface-present) pervasives functions to safestd
;   - made changes to func to that PLAN loading service can use functions
;     directly
;
;
; Revision 1.1  1997/12/05  19:21:01  mwh
; Modified log stuff to have a privileged and an unprivileged interface.
;
 *)
