(* ocamlc -c -nopervasives ieee.ml *)

open Safestd
open Safeunix
open Unixnet
open Safethread
open Log
open Loudbob (* formatting stuff *)
open Loudispan

let ieee_addr=address_to_string ieee_broadcast
type eaddr = int * int
let hostname = ref ""

let value = ref None

let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v


let delimiter =
  "\239\190\173\222\206\250\173\222\239\190\173\222\206\250\173\222"
let multicast_bit = 0x010000		(*bit from high word for multicast*)
let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()
let suspend = ref false
let suspend_mutex = Mutex.create ()
let suspend_condition = Condition.create ()
let hash_timeout = 300			(* 5 minutes *)


(* Caution: this routine returns the address in a byte order that allows
   intercomparison but which would be different from machine to machine *)
let get_eaddr s ofs =
  let a = Char.code (String.get s ofs) in
  let b = Char.code (String.get s (ofs + 1)) in
  let c = Char.code (String.get s (ofs + 2)) in
  let d = Char.code (String.get s (ofs + 3)) in
  let e = Char.code (String.get s (ofs + 4)) in
  let f = Char.code (String.get s (ofs + 5)) in
  ((a lsl 16) lor (b lsl 8) lor c, (d lsl 16) lor (e lsl 8) lor f)

let get_src_eaddr s = get_eaddr s 6
let get_dst_eaddr s = get_eaddr s 0

let get_dsap s = Char.code (String.get s 14)
let get_ssap s = Char.code (String.get s 15)

let get_proto1 s = Char.code (String.get s 12)
let get_proto2 s = Char.code (String.get s 13)

let format_saddr_pkt saddr =
  match saddr with
    ADDR_UNIX s -> "family AF_UNIX path '" ^ s ^ "'"
  | ADDR_INET (a, port) -> "family AF_INET addr " ^ string_of_inet_addr a ^ " port "
      ^ string_of_int port
  | ADDR_PKT (fam, dev, prot) -> "family PACKET dev type " ^ string_of_int fam ^
      " device " ^ dev ^ " proto " ^ string_of_int prot

(* ******************************************************************** *)
let generate_it s =
    let len = 60 in (* (String.length s)-1 in *) 
    let str = ref "String length: " in
    let a= Printf.sprintf "%d\n" len in 

    str := (!str ^ a);
    for i=0 to (len-1) do
      let tmp = Char.code (String.get s i) in
      let stmp = Printf.sprintf "%x " tmp in
      str := (!str ^ stmp);
    done;
    str := (!str ^ "\n");
    !str

(* ******************************************************************** *)
let get_portno saddr =
  (match saddr with
    ADDR_PKT(_, dev, _) ->
      let fred = ref 0 in
      let rec looking j t =
	if (j <= t) then begin
	  let mp = Array.get !mapping j in
	  if(dev = mp.dev) then begin
	    fred := j;
	  end
	  else looking (j+1) t; 
      	end in
      	looking 1 ((Array.length !mapping)-1);
	!fred
  | _ -> raise Exit)


(* ******************************************************************** *)
let spanning_core iport =
  let s =
    Printf.sprintf "spanning started \n" in
  log_msg s;
  init_all_arrays ();
  initialization !hostname;
  while not !die do
    let pkt = get_next_pkt_in iport in
    let dst = get_dst_eaddr pkt.pkt in
    let src = get_src_eaddr pkt.pkt in
    let port_no = get_portno pkt.addr in
    let curtime = time () in
    let (dst1, dst2) = dst in
    let (src1, src2) = src in
    let s =
      Printf.sprintf "spanning: %06x.%06x -> %06x.%06x\n" src1 src2 dst1 dst2
    in
    log_msg s;
    
    let ityp = Char.code (String.get pkt.pkt 20) in
     
    if (ityp = ieee_tcn_bpdu_type) then begin
      let tmp = format_tcn_bpdu pkt.pkt in
      let s=print_tcn tmp in
      log_msg (print_tcn tmp);
      received_tcn_bpdu port_no tmp;
    end;
    if(ityp = ieee_config_bpdu_type) then begin
      let tmp = format_config_bpdu pkt.pkt in
      log_msg (print_config tmp);
      received_config_bpdu port_no tmp; 
      log_msg "RECEIVED CONFIG BPDU\n";
      debug_print "\n ***** RECEIVED CONFIG BPDU *****\n";
    end;
  done;
  let s =
    Printf.sprintf "spanning stopped\n"  in
  log_msg s

let spanning () =
  let iport = bind_addr ieee_addr in
  spanning_core iport


let ieee_do_start iport =
  Safethread.create (hashval()) spanning_core iport


let do_suspend () =
  Mutex.lock suspend_mutex;
  suspend := true;
  die := true;
  Condition.signal die_condition;
  Mutex.unlock suspend_mutex

let do_kill () =
  Mutex.lock suspend_mutex;
  suspend := false;
  Condition.signal suspend_condition;
  Mutex.unlock suspend_mutex

let ieeebridge hn =
  Mutex.lock suspend_mutex;
  while !suspend do
    Condition.wait suspend_condition suspend_mutex
  done;
  Mutex.unlock suspend_mutex;
  hostname := hn;
  let t = Safethread.create (hashval()) spanning () in
  Safethread.join t;
  log_msg "KILLED\n";
  ""

let _ = value := Some(Func.register "ieeebridge" ieeebridge)


