module Func = struct
type operation =
    LOAD of string
  | LOADS of string
  | UNLOAD of string
  | EXEC of string * string

let func_table = Hashtbl.create 257
let func_table_m = Mutex.create ()

let register name func =
  (try
    Mutex.lock func_table_m;
    Hashtbl.remove func_table name;
    Hashtbl.add func_table name func
  with e -> Mutex.unlock func_table_m; raise e);
  Mutex.unlock func_table_m;
  Safethread.magic_create ()

let load_string s =
  let rec aux n =
    try
      let fname = Filename.temp_file "loader" ".cmo" in
      let oc = open_out_bin fname in
      output_string oc s;
      close_out oc;
      Dynlink.loadfile fname
    with Sys_error s -> if n > 0 then aux (n - 1) else raise (Sys_error s) in
  aux 3

let unregister name =
  (try
    Mutex.lock func_table_m;
    Hashtbl.remove func_table name
  with e -> Mutex.unlock func_table_m; raise e);
  Mutex.unlock func_table_m

let voiddammit a =
  a

let call_func (name, arg) =
  try
    Mutex.lock func_table_m;
    let func = Hashtbl.find func_table name in
    Mutex.unlock func_table_m;
    let result = func arg in
    (*print_endline result; flush Pervasives.stdout*)
    voiddammit ()
  with Not_found -> (Mutex.unlock func_table_m;
		     prerr_endline "Unknown function")
  | e -> (Mutex.unlock func_table_m; raise e)

let do_op op =
  match op with
    LOAD(file) ->
      begin try
	Dynlink.loadfile file;
(*	prerr_endline ("loaded " ^ file)*)
	()
      with 
	Sys_error(s) -> prerr_endline ("Sys_error: " ^ s)
      | Dynlink.Error e -> prerr_endline (Dynlink.error_message e)
      end

  | LOADS(str) ->
      begin try
	Dynlink.loadstring str;
	(*load_string str;*)
	(*prerr_endline ("loaded string")*)
	()
      with 
	Sys_error(s) -> prerr_endline ("Sys_error: " ^ s)
      | Dynlink.Error e -> prerr_endline (Dynlink.error_message e)
      end

  | EXEC(func, arg) ->
(* no need for this -- Thread.create does this already
   Thread.create (fun () -> Printexc.catch call_func (func, arg)) (); *)
(*      if func = "do_report" then Dynlink.dump_report (int_of_string arg)
      else begin*)
      Thread.create call_func (func, arg);
      (*let s = Printf.sprintf "started %s" func in
      prerr_endline s*)
      ()
(*      end*)

  | UNLOAD(func) ->
      unregister func(*; prerr_endline ("unregistered " ^ func)*)
end
