(* ocamlc -thread -custom unix.cma threads.cma timer.ml -cclib -lunix 
   -cclib -lthreads *)
(* open Sys
open Unix
*)
open Safeloader
open Safeunix
open Unixnet
open Safethread

type elt = float * (unit -> unit) * Safethread.t

let tvalue = ref None

let hashval() =
  match !tvalue with
    None -> failwith "hash tvalue not set!  (This can't happen.)"
  | Some(v) -> v

let timer_m = Mutex.create ()
let timer_q = ref ([]: elt list)

let sorter a b =
  let (a1,_,_) = a in
  let (b1,_,_) = b in
  a1 < b1

let rec rm_it q a kill_p =
  match q with
    [] -> []
  | h::t ->
      let (h_time,h_func,h_tid) = h in
      let (tim, func) = a in
      if (tim = h_time) & (func == h_func) then begin
	if kill_p then Safethread.kill h_tid;
	rm_it t a kill_p
      end
      else
      	h :: (rm_it t a kill_p)

let remove_it q m kill_p a =
  Mutex.lock m;
  q := rm_it !q a kill_p;
  Mutex.unlock m

let remove = remove_it timer_q timer_m true
let remove_no_kill = remove_it timer_q timer_m false

let add_it q m a =
  let (t,f) = a in
  let new_f () =
    let tim = t -. gettimeofday() in
    delay tim;
    f ();
    remove_no_kill a in
  Mutex.lock m;
  let tid = Safethread.create (hashval()) new_f () in
  if List.length !q = 0 then
    q := [(t,f,tid)]
  else
    q := Sort.merge sorter [(t,f,tid)] !q;
  Mutex.unlock m

let add = add_it timer_q timer_m

let destroy () =
  let rec empty_queue q =
    match q with
      [] -> []
    | (_, _, tid)::t ->
	Safethread.kill tid;
	empty_queue t in
  timer_q := empty_queue !timer_q

let create () = ()

let stub ignored =
  create ();
  ""

let _ =  tvalue := Some(Func.register "timer" stub)
