(* This code performs a function similar to traceroute.  When
 * it arrives at a given machine the machine does the following:
 *
 * if (outgoing) & ((myAddress != finish & hopcnt = 1) | (myAddress = finish) )
 * then
 * 		set outgoing false
 *      set reflect = myAddress
 *      forward packet back towards start
 *
 * elseif (outgoing) & myAddress != finish & hopcnt > 1
 * then
 *		decrement hopcnt by 1
 *      forward packet back towards finish
 *	
 * elseif !outgoing & (myAddress != start)
 * then
 *      forward packet back towards start
 *	
 * else 
 *		display difference between myTime and timestamp
 *		display reflect address
 *
 *)

(* to run this in camlight type 'include <filename>' at the
 * # prompt, where filename = fn of a file named fn.ml.
 *)

open Printf;; (* this loads the printf module *)

let getAddress () = "myAddress";;

type tr_packet = {	outgoing: bool;
		 	start:string;
			finish:string;
			timestamp:int;
			hopcnt:int;
			reflect:string
		};;

let send_wf data display = 
	if (data.outgoing) then
		printf "%s:\n\t{outgoing=true; start=%s; finish=%s; reflect=%s; \
				hopcnt=%d; timestamp=%d}\n"
			    display data.start data.finish data.reflect 
				data.hopcnt data.timestamp
	else
		printf "%s:\n\t{outgoing=false; start=%s; finish=%s; reflect=%s; \
				hopcnt=%d; timestamp=%d}\n"
			    display data.start data.finish data.reflect 
				data.hopcnt data.timestamp
;;

let traceroute tr_packet = 

	begin

	if (tr_packet.outgoing
	    & 
	   ( (getAddress() != tr_packet.finish & tr_packet.hopcnt = 1) ||
             (getAddress() = tr_packet.finish)
           )		
        )		
	then
		send_wf {outgoing = false; 
				 start = tr_packet.start;
				 finish = tr_packet.finish;
				 reflect = getAddress();
				 hopcnt = 0;
				 timestamp = tr_packet.timestamp}
				 "Switchlet has been reflected";

	if (tr_packet.outgoing
	    & 
	   ( (getAddress() != tr_packet.finish & tr_packet.hopcnt > 1) 
             ||
             (getAddress () = tr_packet.finish)
           )
        )		
	then
		send_wf {outgoing = true; 
				 start = tr_packet.start;
				 finish = tr_packet.finish;
				 reflect = getAddress();
				 hopcnt = tr_packet.hopcnt-1;
				 timestamp = tr_packet.timestamp}
				 "Switchlet has been forward toward finish";

	if (tr_packet.outgoing = false & getAddress() != tr_packet.start)
	then
		send_wf {outgoing = tr_packet.outgoing; 
				 start = tr_packet.start;
				 finish = tr_packet.finish;
				 reflect = tr_packet.reflect;
				 hopcnt = tr_packet.hopcnt;
				 timestamp = tr_packet.timestamp}
				 "Switchlet (reflected) has been forward toward start";

	if (tr_packet.outgoing = false & getAddress() = tr_packet.start)
	then
		printf "Reflected From %s\n" tr_packet.reflect;

	end;
;;
	

traceroute {outgoing=true; hopcnt = 1; reflect = "nil"; timestamp=7; finish="him"; start="me"};;

traceroute {outgoing=true; hopcnt = 2; reflect = "nil"; timestamp=7; finish="him"; start="me"};;

traceroute {outgoing=false; hopcnt = 1; reflect = "him"; timestamp=7; finish="him"; start="me"};;
