open Planport
open Safestd
open Safethread
open Log
open Queue
open Mutex
open Condition

(* For threads management *)
let die = ref false
let die_mutex = Mutex.create ()
let die_condition = Condition.create ()

(* Cookie used to get threads from the system *)
let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

(* Packet queue --> everything we get from the various threads goes on 
   this bad boy *)
let packetQ:string Queue.t = Queue.create ()
let packetQ_mutex = Mutex.create ()
let packetQ_cond = Condition.create ()

let queue_packet pkt =
  Mutex.lock packetQ_mutex;
  Queue.add pkt packetQ;
  Condition.signal packetQ_cond;
  Mutex.unlock packetQ_mutex

let dequeue_packet () =
  Mutex.lock packetQ_mutex;
  while (Queue.length packetQ = 0) do
    Condition.wait packetQ_cond packetQ_mutex
  done;
  let pkt = Queue.take packetQ in
  Mutex.unlock packetQ_mutex;
  pkt

(* Thread to suck packets off of the packet Q *)
let rec service_packetQ () =
  let s = dequeue_packet() in
  Log.log_msg ("Got message "^s^"\n");
  service_packetQ()

(* List of I/O threads/ports we currently know about *)
let portsL:((Planport.plan_port * Safethread.t) list) ref = ref []
let portsL_mutex = Mutex.create ()

(* Thread to wait for data from local connection *)
let rec service_local_port p =
  try
    let s = Planport.get_packet p in
    queue_packet s;
    service_local_port p
  with 
    PortOpFailed ->
      service_local_port p
  | ClosedByPeer ->
      Mutex.lock portsL_mutex;
      portsL :=
	 List.fold_left 
	   (fun x y -> 
	     match y with (p2,t) -> if p = p2 then x else y::x)
	   [] !portsL;
      let len = List.length !portsL in
      Mutex.unlock portsL_mutex;
      let ebuf = Printf.sprintf "Closed port (total = %d).\n" len in
      Log.log_msg ebuf

(* Thread to grab local connections *)
let wait_for_local_ports () =
  while not !die do
    let p = Planport.get_local_port () in
    let t = Safethread.create (hashval()) service_local_port p in
    Mutex.lock portsL_mutex;
    portsL := ((p,t) :: !portsL);
    let len = List.length !portsL in
    Mutex.unlock portsL_mutex;
    let ebuf = 
      Printf.sprintf "Spawned thread to watch port (total = %d)\n" len
    in
    Log.log_msg ebuf
  done

(* This will start the process of waiting for input *)
let top_thread () =
  let t = Safethread.create (hashval()) wait_for_local_ports () in 
  let t2 = Safethread.create (hashval()) service_packetQ () in 
  let rec kill_all_threads pid_list =
    match pid_list with
      [] -> ()
    | h :: t -> Safethread.kill h; kill_all_threads t in
  Mutex.lock die_mutex;
  while not !die do
    Condition.wait die_condition die_mutex
  done;
  Mutex.lock portsL_mutex;
  let (_,tl) = List.split !portsL in
  portsL := [];
  Mutex.unlock portsL_mutex;
  kill_all_threads (t::t2::tl)

let pland ignored =
  let t = Safethread.create (hashval()) top_thread () in
  Safethread.join t;
  ""

let _ = value := Some(Func.register "pland" pland)




