open Safestd
open Safeunix
open Unixnet
open Log
open Ip

let value = ref None
let hashval() =
  match !value with
    None -> failwith "hash value not set!  (This can't happen.)"
  | Some(v) -> v

let dump_packet pkt =
  let dump_ip_addr addr =
    let (a, b, c, d) = addr in
    Printf.sprintf "%d.%d.%d.%d" a b c d in
  let dump_ip_hdr ip_hdr =
    Printf.printf "version: %d hdr len: %d tos: %d length: %d\n"
    ip_hdr.version ip_hdr.ihl ip_hdr.tos ip_hdr.length;
    Printf.printf "ident: %04x flags: %x frag offset: %d\n"
    ip_hdr.id ip_hdr.flags ip_hdr.frag_offset;
    Printf.printf "ttl: %d protocol: %d header chksum: %04x\n"
    ip_hdr.ttl ip_hdr.protocol ip_hdr.hdr_chksum;
    Printf.printf "%s -> %s\n"
    (dump_ip_addr ip_hdr.source_addr) (dump_ip_addr ip_hdr.dest_addr) in
  let rec dump_data s ofs len =
    if len >= 16 then begin
      Printf.printf
      "\t%02x %02x %02x %02x  %02x %02x %02x %02x  %02x %02x %02x %02x  %02x %02x %02x %02x  \n"
      (Char.code (String.get s ofs))
      (Char.code (String.get s (ofs + 1)))
      (Char.code (String.get s (ofs + 2)))
      (Char.code (String.get s (ofs + 3)))
      (Char.code (String.get s (ofs + 4)))
      (Char.code (String.get s (ofs + 5)))
      (Char.code (String.get s (ofs + 6)))
      (Char.code (String.get s (ofs + 7)))
      (Char.code (String.get s (ofs + 8)))
      (Char.code (String.get s (ofs + 9)))
      (Char.code (String.get s (ofs + 10)))
      (Char.code (String.get s (ofs + 11)))
      (Char.code (String.get s (ofs + 12)))
      (Char.code (String.get s (ofs + 13)))
      (Char.code (String.get s (ofs + 14)))
      (Char.code (String.get s (ofs + 15)));
      dump_data s (ofs + 16) (len - 16)
    end else begin
      for i = 0 to len - 1 do
	if i = 0 then Printf.printf "\t"
	else if i mod 4 = 0 then Printf.printf "  "
	else Printf.printf " ";
	Printf.printf "%02x" (Char.code (String.get s (ofs + i)))
      done;
      Printf.printf "\n"
    end in
  dump_ip_hdr pkt.ip_hdr;
  Printf.printf "\n";
  dump_data pkt.ip_data 0 (String.length pkt.ip_data);
  Printf.printf "\n\n"

let tstip ignored =
  begin
    try
      let ip_port = bind_ip_proto 17 in
      for i = 1 to 1000 do
      	let pkt = get_packet ip_port in
      	dump_packet pkt
      done;
      unbind_ip_proto ip_port
    with Unix_error(e, s1, s2) ->
      Printf.printf "tstip: unix error: %s '%s'; '%s'\n" (error_message e) s1 s2
    | e -> Printf.printf "tstip: unknown exception\n"
  end;
  ""

let _ = value := Some(Func.register "tstip" tstip)
