type port = PORT of int * Unix.file_descr

type hdr = { src_port : int;
	     dst_port : int;
	     src_addr : Unix.inet_addr;
	     dst_addr : Unix.inet_addr
	   }

exception Already_bound

let port_hashtable = Hashtbl.create 257
let port_hash_m = Mutex.create ()

let bind_proto port =
Printf.printf "bind_proto %d\n" port;
Pervasives.flush Pervasives.stdout;
  try
    Mutex.lock port_hash_m;
    Hashtbl.find port_hashtable port;
    Mutex.unlock port_hash_m;
Printf.printf "bind_proto; Already_bound\n";
Pervasives.flush Pervasives.stdout;
    raise Already_bound
  with Not_found -> begin
    let skt = ThreadUnix.socket Unix.PF_INET Unix.SOCK_DGRAM 0 in
(*Printf.printf "bind_proto; Unix.bind %d\n" port;
Pervasives.flush Pervasives.stdout;*)
    begin
      try
	Unix.bind skt (Unix.ADDR_INET(Unix.inet_addr_any, port))
      with Unix.Unix_error(_) ->
	Mutex.unlock port_hash_m;
	Unix.close skt;
	raise Already_bound
    end;
    Unix.set_nonblock skt;
    Hashtbl.add port_hashtable port skt;
    Mutex.unlock port_hash_m;
    PORT(port, skt)
  end

let unbind_proto p =
  let PORT(port, skt) = p in
  try
    Mutex.lock port_hash_m;
    Hashtbl.remove port_hashtable port;
    Unix.close skt;
    Mutex.unlock port_hash_m
  with _ -> Mutex.unlock port_hash_m

let str_sz = 64*1024

let get_packet p =
  let PORT(dst_port, skt) = p in
  let data = String.create str_sz in
  (* assumes that the file descriptor is set to non-blocking *)
  let (len, sa) =
    (try
      Unix.recvfrom skt data 0 str_sz []
    with Unix.Unix_error(e,w,s) ->
      (match e with
	Unix.EAGAIN ->
	  ThreadUnix.recvfrom skt data 0 str_sz []
      | _ -> raise (Unix.Unix_error(e,w,s)))) in
  match sa with
    Unix.ADDR_INET(src_addr, src_port) ->
      let hdr = { src_port=src_port;
		  dst_port=dst_port;
		  src_addr=src_addr;
		  dst_addr=Unix.inet_addr_any
		} in
      (hdr, len, data)
  | _ -> raise (Failure "get_packet")

let send uport hdr payload =
  let PORT(src_port, skt) = uport in
  if hdr.src_port <> src_port then begin
(*Printf.printf "hdr.src_port (%d) <> src_port (%d)\n" hdr.src_port src_port;
Pervasives.flush Pervasives.stdout;*)
    raise (Failure "Udpsock.send")
  end else ThreadUnix.sendto skt payload 0 (String.length payload) []
      (Unix.ADDR_INET(hdr.dst_addr, hdr.dst_port))
      
let reply uport hdr payload =
  let PORT(src_port, skt) = uport in
  if hdr.dst_port <> src_port then
    raise (Failure "Udpsock.send")
  else ThreadUnix.sendto skt payload 0 (String.length payload) []
      (Unix.ADDR_INET(hdr.src_addr, hdr.src_port))

let get_source hdr = hdr.src_addr
let get_dest hdr = hdr.dst_addr
