open Safeloader
open Printf
open Wf
open Safeunix
open Log
open An_marshal
open Saneproto
(*open Time*)

(*let _ = start_time 6;;*)

type ping_packet = { start : string ; finish : string; timestamp : float}

let ping_encode pkt =
  Printf.sprintf "start = %s; finish = %s; timestamp = %f"
    pkt.start pkt.finish pkt.timestamp

let decode_regexp =
  Str.regexp "^start = \(.*\); finish = \(.*\); timestamp = \(.*\)$"
let ping_decode str ofs =
  if not (Str.string_match decode_regexp str ofs)
  then failwith "bad packet"
  else
    { start = Str.matched_group 1 str;
      finish = Str.matched_group 2 str;
      timestamp = float_of_string (Str.matched_group 3 str)
    }

(*
 * This is routine that starts things off by handing the components
 * of an ANEP to send_wf.  We took one of Nettles Type_IDs = 20
 *)
let send_ping dest name = 
  let pingcode = Get_bytecode.get_bytecode "saneping" in
  send_sane dest (Route.get_route dest) {do_forward = true; type_id = 20}  []
    "ping_out" pingcode
    (ping_encode { start = An.getAddress ();
  		   finish = dest;
  		   timestamp = Time.get_time ()
		 })

let ping_out arg_string =
  let {code=code; data=datastr; func=func} = decode arg_string 0 in
  let ping_packet = ping_decode datastr 0 in
  if  (ping_packet.finish = An.getAddress())
  then begin
    send_sane (ping_packet.start) (Route.get_route ping_packet.start)
      {do_forward=true; type_id=20} [] "ping_in" code datastr;
(*    let str =
      "at the remote machine so set outgoing false and send toward start" in
    Log.log_msg (str ^ "\n"); str*) ""
  end else begin
    send_sane (ping_packet.finish) (Route.get_route ping_packet.finish)
      {do_forward=true; type_id=20} [] "ping_out" code datastr;
(*    let str =
      "not there yet so forward the packet toward finish" in
    Log.log_msg (str ^ "\n"); str *) ""
  end

let ping_in arg_string =
  let {code=code; data=datastr; func=func} = decode arg_string 0 in
  let ping_packet = ping_decode datastr 0 in
    if (ping_packet.start = An.getAddress() )
    then begin
      let tot_time = (Time.get_time()) -. ping_packet.timestamp in
      log_msg
	  (Printf.sprintf "back at the sending machine .. Success (%f sec)\n"
	     tot_time);
(*      let str =  "back at the sending machine .. Success" in
      Log.log_msg (str ^ "\n"); str*) ""
    end else begin
      send_sane (ping_packet.start) (Route.get_route ping_packet.start)
	{ do_forward=true; type_id=20 } [] "ping_in" code datastr;
      (*let str = "not back yet so forward the packet toward start" in
      Log.log_msg (str ^ "\n"); str *) ""
    end

(*let _ = end_time 6

let _ = start_time 7*)
let _ = Func.register "ping_out" ping_out
let _ = Func.register "ping_in" ping_in
(*let _ = end_time 7*)
