open Safeloader

type sha1_ctx = { state : Int32.int32 array;(*len = 5*)
		  count : int array;	(*len = 2*)
		  buffer : string }	(*len = 64*)

let ctx_copy ctx =
  { state = Array.copy ctx.state;
    count = Array.copy ctx.count;
    buffer = String.copy ctx.buffer
  }

let set_int32 (a,b) =
  Int32.(lor) (Int32.(lsl) (Int32.int32_of_int a) 16) (Int32.int32_of_int b)

let x00ff00ff = set_int32 (0x00ff, 0x00ff)
let x10325476 = set_int32 (0x1032, 0x5476)
let x5a827999 = set_int32 (0x5a82, 0x7999)
let x67452301 = set_int32 (0x6745, 0x2301)
let x6ed9eba1 = set_int32 (0x6ed9, 0xeba1)
let x8f1bbcdc = set_int32 (0x8f1b, 0xbcdc)
let x98badcfe = set_int32 (0x98ba, 0xdcfe)
let xc3d2e1f0 = set_int32 (0xc3d2, 0xe1f0)
let xca62c1d6 = set_int32 (0xca62, 0xc1d6)
let xefcdab89 = set_int32 (0xefcd, 0xab89)
let xff00ff00 = set_int32 (0xff00, 0xff00)
let xffffffff = set_int32 (0xffff, 0xffff)

let rol32 v len =
  Int32.(land) (Int32.(lor) (Int32.(lsr) v (32 - len)) (Int32.(lsl) v len))
    xffffffff

let rol132 v = rol32 v 1
let rol532 v = rol32 v 5
let rol3032 v = rol32 v 30
let lxor432 a b c d = Int32.(lxor) (Int32.(lxor) a b) (Int32.(lxor) c d)

let sha1transform state buffer =
  let chars_to_longs chars =
    let char4_to_long i =
      (Int32.(lor) (Int32.(lor) (Int32.(lor) (Int32.int32_of_int
					  (Char.code chars.[i]))
 			       (Int32.(lsl) (Int32.int32_of_int
					     (Char.code chars.[i+1])) 8))
 	 	    (Int32.(lsl) (Int32.int32_of_int
				  (Char.code chars.[i+2])) 16))
 	 (Int32.(lsl) (Int32.int32_of_int (Char.code chars.[i+3])) 24)) in
    let longs = Array.create 16 Int32.min_int32 in
    for i = 0 to 15 do
      longs.(i) <- char4_to_long (i*4)
    done;
    longs in
  let longs_to_chars longs chars =
    let long_to_4chars i =
      chars.[i*4] <-
	 Char.chr ((Int32.int_of_int32 longs.(i)) land 0xff);
      chars.[i*4+1] <-
	 Char.chr ((Int32.int_of_int32 (Int32.(lsr) longs.(i) 8)) land 0xff);
      chars.[i*4+2] <-
	 Char.chr ((Int32.int_of_int32 (Int32.(lsr) longs.(i) 16)) land 0xff);
      chars.[i*4+3] <-
	 Char.chr (Int32.int_of_int32 (Int32.(lsr) longs.(i) 24)) in
    for i = 0 to 15 do
      long_to_4chars i
    done in
(*  let rol value bits = Int32.(lor) (Int32.(lsl) value bits) (Int32.(lsr) value (32 - bits)) in*)
  let blk0 b i =			(* assume little-endian *)
    b.(i) <- Int32.(lor) (Int32.(land) (rol32 b.(i) 24) xff00ff00)
      (Int32.(land) (rol32 b.(i) 8) x00ff00ff);
    b.(i) in
  let blk b i =
    b.(i land 15) <- rol132 (lxor432 b.((i+13)land 15) b.((i+8) land 15)
			       b.((i+2) land 15) b.(i land 15));
    b.(i land 15) in
  let r0 b s v w x y z i =
    s.(z) <-
       Int32.(land)
       (Int32.add s.(z)
	  (Int32.add (Int32.(lxor) (Int32.(land) s.(w) (Int32.(lxor) s.(x) s.(y))) s.(y))
	     (Int32.add (blk0 b i)
		(Int32.add x5a827999 (rol532 s.(v))))))
	 xffffffff;
    s.(w) <- rol3032 s.(w) in
  let r1 b s v w x y z i =
    s.(z) <-
       Int32.(land)
       (Int32.add s.(z) 
	  (Int32.add (Int32.(lxor) (Int32.(land) s.(w) (Int32.(lxor) s.(x) s.(y))) s.(y))
	     (Int32.add (blk b i)
		(Int32.add x5a827999 (rol532 s.(v))))))
	 xffffffff;
    s.(w) <- rol3032 s.(w) in
  let r2 b s v w x y z i =
    s.(z) <-
       Int32.(land)
       (Int32.add s.(z)
	  (Int32.add (Int32.(lxor) s.(w) (Int32.(lxor) s.(x) s.(y)))
	     (Int32.add (blk b i)
		(Int32.add x6ed9eba1 (rol532 s.(v))))))
	 xffffffff;
    s.(w) <- rol3032 s.(w)in
  let r3 b s v w x y z i =
    s.(z) <-
       Int32.(land)
       (Int32.add s.(z)
	  (Int32.add (Int32.(lor) (Int32.(land) (Int32.(lor) s.(w) s.(x)) s.(y)) (Int32.(land) s.(w) s.(x)))
	     (Int32.add (blk b i)
		(Int32.add x8f1bbcdc (rol532 s.(v))))))
	 xffffffff;
    s.(w) <- rol3032 s.(w) in
  let r4 b s v w x y z i =
    s.(z) <-
       Int32.(land)
       (Int32.add s.(z)
	  (Int32.add (Int32.(lxor) s.(w) (Int32.(lxor) s.(x) s.(y)))
	     (Int32.add (blk b i)
		(Int32.add xca62c1d6 (rol532 s.(v))))))
	 xffffffff;
    s.(w) <- rol3032 s.(w) in
  let block = chars_to_longs buffer in
    (* 4 rounds of 20 operations each. Loop unrolled. *)
  let tmp = Array.copy state in
  r0 block tmp 0 1 2 3 4 0;
  r0 block tmp 4 0 1 2 3 1;
  r0 block tmp 3 4 0 1 2 2;
  r0 block tmp 2 3 4 0 1 3;
  r0 block tmp 1 2 3 4 0 4;
  r0 block tmp 0 1 2 3 4 5;
  r0 block tmp 4 0 1 2 3 6;
  r0 block tmp 3 4 0 1 2 7;
  r0 block tmp 2 3 4 0 1 8;
  r0 block tmp 1 2 3 4 0 9;
  r0 block tmp 0 1 2 3 4 10;
  r0 block tmp 4 0 1 2 3 11;
  r0 block tmp 3 4 0 1 2 12;
  r0 block tmp 2 3 4 0 1 13;
  r0 block tmp 1 2 3 4 0 14;
  r0 block tmp 0 1 2 3 4 15;
  r1 block tmp 4 0 1 2 3 16;
  r1 block tmp 3 4 0 1 2 17;
  r1 block tmp 2 3 4 0 1 18;
  r1 block tmp 1 2 3 4 0 19;
  r2 block tmp 0 1 2 3 4 20;
  r2 block tmp 4 0 1 2 3 21;
  r2 block tmp 3 4 0 1 2 22;
  r2 block tmp 2 3 4 0 1 23;
  r2 block tmp 1 2 3 4 0 24;
  r2 block tmp 0 1 2 3 4 25;
  r2 block tmp 4 0 1 2 3 26;
  r2 block tmp 3 4 0 1 2 27;
  r2 block tmp 2 3 4 0 1 28;
  r2 block tmp 1 2 3 4 0 29;
  r2 block tmp 0 1 2 3 4 30;
  r2 block tmp 4 0 1 2 3 31;
  r2 block tmp 3 4 0 1 2 32;
  r2 block tmp 2 3 4 0 1 33;
  r2 block tmp 1 2 3 4 0 34;
  r2 block tmp 0 1 2 3 4 35;
  r2 block tmp 4 0 1 2 3 36;
  r2 block tmp 3 4 0 1 2 37;
  r2 block tmp 2 3 4 0 1 38;
  r2 block tmp 1 2 3 4 0 39;
  r3 block tmp 0 1 2 3 4 40;
  r3 block tmp 4 0 1 2 3 41;
  r3 block tmp 3 4 0 1 2 42;
  r3 block tmp 2 3 4 0 1 43;
  r3 block tmp 1 2 3 4 0 44;
  r3 block tmp 0 1 2 3 4 45;
  r3 block tmp 4 0 1 2 3 46;
  r3 block tmp 3 4 0 1 2 47;
  r3 block tmp 2 3 4 0 1 48;
  r3 block tmp 1 2 3 4 0 49;
  r3 block tmp 0 1 2 3 4 50;
  r3 block tmp 4 0 1 2 3 51;
  r3 block tmp 3 4 0 1 2 52;
  r3 block tmp 2 3 4 0 1 53;
  r3 block tmp 1 2 3 4 0 54;
  r3 block tmp 0 1 2 3 4 55;
  r3 block tmp 4 0 1 2 3 56;
  r3 block tmp 3 4 0 1 2 57;
  r3 block tmp 2 3 4 0 1 58;
  r3 block tmp 1 2 3 4 0 59;
  r4 block tmp 0 1 2 3 4 60;
  r4 block tmp 4 0 1 2 3 61;
  r4 block tmp 3 4 0 1 2 62;
  r4 block tmp 2 3 4 0 1 63;
  r4 block tmp 1 2 3 4 0 64;
  r4 block tmp 0 1 2 3 4 65;
  r4 block tmp 4 0 1 2 3 66;
  r4 block tmp 3 4 0 1 2 67;
  r4 block tmp 2 3 4 0 1 68;
  r4 block tmp 1 2 3 4 0 69;
  r4 block tmp 0 1 2 3 4 70;
  r4 block tmp 4 0 1 2 3 71;
  r4 block tmp 3 4 0 1 2 72;
  r4 block tmp 2 3 4 0 1 73;
  r4 block tmp 1 2 3 4 0 74;
  r4 block tmp 0 1 2 3 4 75;
  r4 block tmp 4 0 1 2 3 76;
  r4 block tmp 3 4 0 1 2 77;
  r4 block tmp 2 3 4 0 1 78;
  r4 block tmp 1 2 3 4 0 79;
  for i = 0 to 4 do
    state.(i) <- Int32.(land) (Int32.add state.(i) tmp.(i)) xffffffff
  done

let sha1init () =
  let ctx = { state = Array.create 5 x67452301;
	      count = Array.create 2 0;
	      buffer = String.make 64 ' ' } in
  ctx.state.(1) <- xefcdab89;
  ctx.state.(2) <- x98badcfe;
  ctx.state.(3) <- x10325476;
  ctx.state.(4) <- xc3d2e1f0;
  ctx

let sha1update ctx data =
  let j = ref ctx.count.(0) in
  let i = ref 0 in
  let len = String.length data in
  ctx.count.(0) <- ctx.count.(0) + (len lsl 3);
  if ctx.count.(0) < !j then
    ctx.count.(1) <- ctx.count.(1) + ((len lsr 29) + 1);
  j := (!j lsr 3) land 63;
  if (!j + len) > 63 then begin
    i := 64 - !j;
    String.blit data 0 ctx.buffer (!j) (!i);
    sha1transform ctx.state ctx.buffer;
    while (!i + 63) < len do
      String.blit data (!i) ctx.buffer 0 64;
      sha1transform ctx.state ctx.buffer;
      i := !i + 64
    done
  end;
  String.blit data (!i) ctx.buffer !j (len - !i)

let sha1final ctx =
  let digest = String.make 20 ' ' in
  let finalcount = String.make 8 ' ' in
  for i = 0 to 7 do
    let idx = if i > 3 then 0 else 1 in
    finalcount.[i] <-
       Char.chr (ctx.count.(idx) lsr ((3-(i land 3)) * 8) land 255)
  done;
  sha1update ctx "\128";
  while (ctx.count.(0) land 504) <> 448 do
    sha1update ctx "\000"
  done;
  sha1update ctx finalcount;

  for i = 0 to 19 do
    digest.[i] <-
       Char.chr (Int32.int_of_int32 (Int32.(land) (Int32.(lsr) ctx.state.(i lsr 2)
						     ((3 - (i land 3)) * 8))
				       (Int32.int32_of_int 255)))
  done;
  digest
